// StretchTga.cpp: implementation file

#include "stdafx.h"

#include <afxtempl.h>
#include <float.h>

#include <list>
#include "Waypoint.h"

#include "geoDataClient.h"
#include "geoDataClientDialog.h"
#include "HTTPRequest.h"

#include <math.h>

// Stretch/Shrink two TGA files.
// Returns 0 if OK.
int CgeoDataClientDialog::StretchTga(const CString &tgaFilenameInput,
                                     const CString &tgaFilenameOutput,
                                     double x1_, double y1_,
                                     double x2_, double y2_,
                                     double x3_, double y3_,
                                     double x4_, double y4_,
                                     int nWidth, int nHeight)
{
  int     x, y;
  int     xSource, ySource;
  RGBQUAD Color;
  TGAFile tgaFileInput;
  TGAFile tgaFileOutput;

  CString message;
  message.Format(_T("Stretching/shrinking \"%s\" to \"%s\"."),
                 tgaFilenameInput,
                 tgaFilenameOutput);
  addMessage(message);

  double dA[20]; /* N * (N + 1) coefficients         */
  double dX[8];  /* N * 2 unknown in N * 2 equations */

  dA[ 0] =   0;                /* x1      */
  dA[ 1] =   0;                /* y1      */
  dA[ 2] =   0;                /* x1 * y1 */
  dA[ 3] =   1.00;
  dA[ 4] =   x1_;
  dA[ 5] =   nWidth;           /* x2      */
  dA[ 6] =   0;                /* y2      */
  dA[ 7] =   0;                /* x2 * y2 */
  dA[ 8] =   1.00;
  dA[ 9] =   x2_;
  dA[10] =   0;                /* x3      */
  dA[11] =   nHeight;          /* y3      */
  dA[12] =   0;                /* x3 * y3 */
  dA[13] =   1.00;
  dA[14] =   x3_;
  dA[15] =   nWidth;           /* x4      */
  dA[16] =   nHeight;          /* y4      */
  dA[17] =   nWidth * nHeight; /* x4 * y4 */
  dA[18] =   1.00;
  dA[19] =   x4_;

  dX[ 0] =   0.00;
  dX[ 1] =   0.00;
  dX[ 2] =   0.00;
  dX[ 3] =   0.00;

  if(GaussJordan(4, dA, dX)) /* Solve first four equations. */
  {
    addMessage(_T("Error in StretchTga(): Indefinite solution in x."));
    return 1;
  }

  dA[ 0] =   0;                /* x1      */
  dA[ 1] =   0;                /* y1      */
  dA[ 2] =   0;                /* x1 * y1 */
  dA[ 3] =   1.00;
  dA[ 4] =   y1_;
  dA[ 5] =   nWidth;           /* x2      */
  dA[ 6] =   0;                /* y2      */
  dA[ 7] =   0;                /* x2 * y2 */
  dA[ 8] =   1.00;
  dA[ 9] =   y2_;
  dA[10] =   0;                /* x3      */
  dA[11] =   nHeight;          /* y3      */
  dA[12] =   0;                /* x3 * y3 */
  dA[13] =   1.00;
  dA[14] =   y3_;
  dA[15] =   nWidth;           /* x4      */
  dA[16] =   nHeight;          /* y4      */
  dA[17] =   nWidth * nHeight; /* x4 * y4 */
  dA[18] =   1.00;
  dA[19] =   y4_;

  dX[ 4] =   0.00;
  dX[ 5] =   0.00;
  dX[ 6] =   0.00;
  dX[ 7] =   0.00;

  if(GaussJordan(4, dA, dX + 4)) /* Solve next four equations. */
  {
    addMessage(_T("Error in StretchTga(): Indefinite solution in y."));
    return 2;
  }

  if(tgaFileInput.Open(tgaFilenameInput))
  {
    CString message;
    message.Format(_T("Error in StretchTga(): Could not open TGA file \"%s\"."), tgaFilenameInput);
    addMessage(message);
    return 3;
  }
  if(tgaFileOutput.New(nWidth, nHeight))
  {
    tgaFileInput.Close();
    CString message;
    message.Format(_T("Error in StretchTga(): Could not initialize %d (w) by %d (h) TGA file."), nWidth, nHeight);
    addMessage(message);
    return 4;
  }

  for(y = 0; y < nHeight; y++)
  {
    for(x = 0; x < nWidth; x++)
    {
      xSource = int(dX[0] * (x + 0.0) +
                (tgaFileOutput.GetImageHeight() - (y - 0.0)) *
                  (dX[1] + dX[2] * (x + 0.0)) + dX[3] + 0.5);
      ySource = int(tgaFileInput.GetImageHeight() -
                (dX[4] * (x + 0.5) +
                (tgaFileOutput.GetImageHeight() - (y - 0.0)) *
                  (dX[5] + dX[6] * (x + 0.0)) + dX[7]) + 0.5);
      memset(&Color, 0, sizeof(Color));
      tgaFileInput.GetColor(xSource, ySource, &Color);
      tgaFileOutput.SetColor(x, y, &Color);
    }
  }
  if(tgaFileOutput.Write(tgaFilenameOutput))
  {
    CString message;
    message.Format(_T("Error in StretchTga(): Could not write TGA file \"%s\"."), nWidth, nHeight);
    addMessage(message);
    tgaFileOutput.Close();
    tgaFileInput.Close();
    return 5;
  }
  message.Format(_T("Stretching/shrinking successful: \"%s\" written."),
                 tgaFilenameOutput);
  addMessage(message);
  tgaFileOutput.Close();
  tgaFileInput.Close();
  return 0;
}

/*
GaussJordan() solves linear equations.
It returns 1 if a division by a near-zero value
would have been done (indefinite solution).

It returns 0 if the solution is sufficiently definite.
*/
int CgeoDataClientDialog::GaussJordan(int nDimension,
                                      double *pdA,
                                      double *pdX)
{
  /* Example:
                 1.00 * x1 - 2.00 * x2 + 3.00 * x3 =  0.00
                 2.00 * x1 + 5.00 * x2 - 3.00 * x3 = 27.00
                 3.00 * x1 - 2.00 * x2 + 6.00 * x3 =  2.00
    nDimension is the number of unknowns, e. g. 3.
    pdA points to the array of coefficients, e. g.
                  1.00,  -2.00,   3.00,   0.00,
                  2.00,   5.00,  -3.00,  27.00,
                  3.00,  -2.00,   6.00,   2.00
    pDX points to the array of unknowns to be calculated.
    GaussJordan should yield
                 16.00,  -7.00, -10.00

    From: Schauer, Pascal fr Anfnger, dritte Auflage, S. 133
  */
  int nI;
  int nJ;
  double dQ;
  const double epsilon = 1.0e-12;
  int bError = 0;
  for(nI = 0; nI < nDimension - 1; nI++)
  {
    int nM;
    int nK;
    nM = nI;
    for(nK = nI + 1; nK < nDimension; nK++)
    {
      if(fabs(pdA[nK * (nDimension + 1) + nI]) >
         fabs(pdA[nM * (nDimension + 1) + nI]))
      {
        nM = nK;
      }
    }
    if(nM != nI)
    {
      /* Swap lines. */
      for(nJ = nI; nJ < nDimension + 1; nJ++)
      {
        dQ = pdA[nI * (nDimension + 1) + nJ];
        pdA[nI * (nDimension + 1) + nJ] = pdA[nM * (nDimension + 1) + nJ];
        pdA[nM * (nDimension + 1) + nJ] = dQ;
      }
    }
    for(nK = nI + 1; nK < nDimension; nK++)
    {
      double dDivisor;
      dDivisor = pdA[nI * (nDimension + 1) + nI];
      if(fabs(dDivisor) < epsilon)
      {
        bError = 1;
        if(pdA[nK * (nDimension + 1) + nI] == 0.0)
        {
          dQ = 1.0;
        }
        else
        {
          dQ = pdA[nK * (nDimension + 1) + nI] / epsilon;
        }
      }
      else
      {
        dQ = pdA[nK * (nDimension + 1) + nI] / dDivisor;
      }
      for(nJ = nI + 1; nJ < nDimension + 1; nJ++)
      {
        pdA[nK * (nDimension + 1) + nJ] =
          pdA[nK * (nDimension + 1) + nJ] -
          pdA[nI * (nDimension + 1) + nJ] * dQ;
      }
    }
  }
  for(nI = nDimension - 1; nI >= 0; nI--)
  {
    double dDivisor;
    dQ = pdA[nI * (nDimension + 1) + nDimension];
    for(nJ = nI + 1; nJ < nDimension; nJ++)
    {
      dQ = dQ - pdA[nI * (nDimension + 1) + nJ] * pdX[nJ];
    }
    dDivisor = pdA[nI * (nDimension + 1) + nI];
    if(fabs(dDivisor) < epsilon)
    {
      bError = 1;
      if(dQ == 0.0)
      {
        pdX[nI] = 1.0;
      }
      else
      {
        pdX[nI] = dQ / epsilon;
      }
    }
    else
    {
      pdX[nI] = dQ / dDivisor;
    }
  }
  return bError;
}
